import 'dart:developer';

import '../../classes/Devices/bc_device.dart';
import '../../classes/Devices/be_device.dart';
import '../../classes/Devices/bg_device.dart';
import '../../classes/Devices/bp_device.dart';
import '../../classes/Devices/bt_device.dart';
import '../../classes/Devices/dd_device.dart';
import '../../classes/Devices/ms_device.dart';
import '../../classes/Devices/oc_device.dart';
import '../../classes/Devices/pr_device.dart';
import '../../classes/Devices/zi_device.dart';
import '../../classes/Devices/zl_device.dart';
import 'package:xml/xml.dart';
import '../classes.dart';
import '../device.dart';

class AuthenticateResponse extends CuppsBase {
  final String result;
  final String deviceToken;
  final PlatformParameter platformParameter;
  final ApplicationParameter applicationParameter;
  final List<Device> deviceList;
  PlatformDevices platformDevices;

  AuthenticateResponse({
    required int messageId,
    required this.result,
    required this.deviceToken,
    required this.platformParameter,
    required this.applicationParameter,
    required this.deviceList,
    required this.platformDevices,
  }) : super(messageID: 3, messageName: "authenticateResponse");

  factory AuthenticateResponse.fromXML(String xmlSTR) {
    final document = XmlDocument.parse(xmlSTR);

    String messageID = document.rootElement.attributes
        .firstWhere((p0) => p0.name.toString() == "messageID")
        .value;
    String result = document.rootElement.childElements.first.attributes
        .firstWhere((p0) => p0.name.toString() == "result")
        .value;
    String deviceToken = document.rootElement.childElements.first.attributes
        .firstWhere((p0) => p0.name.toString() == "deviceToken")
        .value;

    XmlElement? platformParameterElement =
        document.findAllElements("platformParameter").first;
    XmlElement? applicationParameterElement =
        document.findAllElements("applicationParameter").first;
    XmlElement? deviceListElement =
        document.findAllElements("deviceList").first;
    List<XmlElement> devicesElements =
        deviceListElement.findElements("device").toList();



    PlatformParameter platformParameter =
        PlatformParameter.fromXML(platformParameterElement.toXmlString());
    ApplicationParameter applicationParameter =
        ApplicationParameter.fromXML(applicationParameterElement.toXmlString());
    List<Device> deviceList =
        devicesElements.map((e) => Device.fromXML(e.toXmlString())).toList();

    PlatformDevices platformDevices = PlatformDevices(
        bcDevices: deviceList
            .where((element) => element.getType == DeviceType.bc)
            .map((e) => BcDevice.fromDevice(e))
            .toList(),
      beDevices: deviceList
          .where((element) => element.getType == DeviceType.be)
          .map((e) => BeDevice.fromDevice(e))
          .toList(),
      bgDevices: deviceList
          .where((element) => element.getType == DeviceType.bg)
          .map((e) => BgDevice.fromDevice(e))
          .toList(),
      bpDevices: deviceList
          .where((element) => element.getType == DeviceType.bp)
          .map((e) => BpDevice.fromDevice(e))
          .toList(),
      btDevices: deviceList
          .where((element) => element.getType == DeviceType.bt)
          .map((e) => BtDevice.fromDevice(e))
          .toList(),
      ddDevices:  deviceList
          .where((element) => element.getType == DeviceType.dd)
          .map((e) => DdDevice.fromDevice(e))
          .toList(),
        msDevices: deviceList
            .where((element) => element.getType == DeviceType.ms)
            .map((e) => MsDevice.fromDevice(e))
            .toList(),
        ocDevices: deviceList
            .where((element) => element.getType == DeviceType.oc)
            .map((e) => OcDevice.fromDevice(e))
            .toList(),
        prDevices: deviceList
            .where((element) => element.getType == DeviceType.pr)
            .map((e) => PrDevice.fromDevice(e))
            .toList(),
        ziDevices: deviceList
            .where((element) => element.getType == DeviceType.zi)
            .map((e) => ZiDevice.fromDevice(e))
            .toList(),
        zlDevices: deviceList
            .where((element) => element.getType == DeviceType.zl)
            .map((e) => ZlDevice.fromDevice(e))
            .toList(),
    );

    return AuthenticateResponse(
        result: result,
        deviceToken: deviceToken,
        platformParameter: platformParameter,
        applicationParameter: applicationParameter,
        deviceList: deviceList,
        messageId: int.parse(messageID),
        platformDevices: platformDevices);
  }

  @override
  String toXML() {
    return '''
    <?xml version="1.0" encoding="utf-8"?>
    <cupps xmlns="$xmlns" messageID="$messageID" messageName="$messageName" xmlns:xyzi="$xsi">
      <$messageName result="$result" deviceToken="$deviceToken">
          ${platformParameter.toXML()}
          ${applicationParameter.toXML()}
          <deviceList >
              ${deviceList.map((e) => e.toXML()).join()}
          </deviceList>
      </authenticateResponse>
    </cupps>'''
        .replaceAll("\n", "")
        .replaceAll("\t", "")
        .replaceAll("  ", "");
  }

  @override
  String toString() {
    return "$messageID : $messageName";
  }
}

class PlatformParameter {
  final String platformVendor;
  final String platformVersion;
  final CryptAlgorithmInfo cryptAlgorithmInfo;

  PlatformParameter(
      {required this.platformVendor,
      required this.platformVersion,
      required this.cryptAlgorithmInfo});

  factory PlatformParameter.fromXML(String xmlSTR) {
    final document = XmlDocument.parse(xmlSTR);

    String platformVendor = document.rootElement.attributes
        .firstWhere((p0) => p0.name.toString() == "platformVendor")
        .value;
    String platformVersion = document.rootElement.attributes
        .firstWhere((p0) => p0.name.toString() == "platformVersion")
        .value;

    XmlElement? cryptAlgorithmInfoElement =
        document.firstChild!.getElement("cryptAlgorithmInfo");

    CryptAlgorithmInfo cryptAlgorithmInfo =
        CryptAlgorithmInfo.fromXML(cryptAlgorithmInfoElement!.toString());

    return PlatformParameter(
        platformVendor: platformVendor,
        platformVersion: platformVersion,
        cryptAlgorithmInfo: cryptAlgorithmInfo);
  }

  String toXML() {
    return '''
              <platformParameter platformVendor="$platformVendor" platformVersion="$platformVersion">
                  ${cryptAlgorithmInfo.toXML()}
              </platformParameter>
    '''
        .replaceAll("\n", "")
        .replaceAll("\t", "")
        .replaceAll("  ", "");
  }
}

class CryptAlgorithmInfo {
  final CryptAlgorithm defaultCryptAlgorithm;
  final List<CryptAlgorithm> availableCryptAlgorithms;

  CryptAlgorithmInfo(
      {required this.defaultCryptAlgorithm,
      required this.availableCryptAlgorithms});

  factory CryptAlgorithmInfo.fromXML(String xmlSTR) {
    final document = XmlDocument.parse(xmlSTR);
    XmlElement defaultCryptAlgorithmElement =
        document.rootElement.findAllElements("defaultCryptAlgorithm").first;
    XmlElement cryptAlgorithmListElement =
        document.rootElement.findAllElements("availableCryptAlgorithms").first;
    XmlElement defaultCrypt =
        defaultCryptAlgorithmElement.findElements("cryptAlgorithm").first;
    List<XmlElement> cryptos =
        cryptAlgorithmListElement.findElements("cryptAlgorithm").toList();
    CryptAlgorithm defaultCryptAlgorithm =
        CryptAlgorithm.fromXML(defaultCrypt.toXmlString());
    List<CryptAlgorithm> cryptAlgorithmList =
        cryptos.map((e) => CryptAlgorithm.fromXML(e.toXmlString())).toList();

    return CryptAlgorithmInfo(
        defaultCryptAlgorithm: defaultCryptAlgorithm,
        availableCryptAlgorithms: cryptAlgorithmList);
  }

  String toXML() {
    return '''
              <cryptAlgorithmInfo>
                  <defaultCryptAlgorithm>
                      ${defaultCryptAlgorithm.toXML()}
                  </defaultCryptAlgorithm>
                  <availableCryptAlgorithms>
                      ${availableCryptAlgorithms.map((e) => e.toXML()).join("")}
                  </availableCryptAlgorithms>
              </cryptAlgorithmInfo>
    '''
        .replaceAll("\n", "")
        .replaceAll("\t", "")
        .replaceAll("  ", "");
  }
}

class CryptAlgorithm {
  final String name;

  CryptAlgorithm({required this.name});

  factory CryptAlgorithm.fromXML(String xmlSTR) {
    final document = XmlDocument.parse(xmlSTR);
    String name = document.getAttribute("name") ?? "";

    return CryptAlgorithm(name: name);
  }

  String toXML() {
    return '''<cryptAlgorithm name="$name"/>'''
        .replaceAll("\n", "")
        .replaceAll("\t", "")
        .replaceAll("  ", "");
  }
}

class ApplicationParameter {
  final List<Storage> storages;

  ApplicationParameter({required this.storages});

  factory ApplicationParameter.fromXML(String xmlSTR) {
    final document = XmlDocument.parse(xmlSTR);
    List<XmlElement> storageElements =
        document.rootElement.findAllElements("storage").toList();
    List<Storage> storages =
        storageElements.map((e) => Storage.fromXML(e.toXmlString())).toList();

    return ApplicationParameter(storages: storages);
  }

  String toXML() {
    return '''
    <applicationParameter>
    ${storages.map((e) => e.toXML()).join()}
    </applicationParameter>
    ''';
  }
}

class Storage {
  final String storageT;
  final String path;
  final String driveLetter;

  Storage(
      {required this.storageT, required this.path, required this.driveLetter});

  factory Storage.fromXML(String xmlSTR) {
    final document = XmlDocument.parse(xmlSTR);
    String storageT = document.firstChild!.attributes
        .firstWhere((element) => element.name.toString() == "storageT")
        .value;
    String path = document.firstChild!.attributes
        .firstWhere((element) => element.name.toString() == "path")
        .value;
    String driveLetter = document.firstChild!.attributes
        .firstWhere((element) => element.name.toString() == "driveLetter")
        .value;
    return Storage(storageT: storageT, path: path, driveLetter: driveLetter);
  }

  String toXML() {
    return '''<storage storageT="$storageT" path="$path" driveLetter="$driveLetter"/>''';
  }
}
