import 'dart:developer';

import 'package:artemis_cupps/artemis_cupps.dart';

import '../../classes/Requests/authenticate_req.dart';
import '../../classes/Requests/bye_req.dart';
import '../../classes/Requests/interface_level_req.dart';
import '../../classes/Requests/interface_levels_available_req.dart';
import '../../classes/Responses/application_stop_command_res.dart';
import '../../classes/Responses/authenticate_res.dart';
import '../../classes/Responses/bye_res.dart';
import '../../classes/Responses/interface_level_res.dart';
import '../../classes/Responses/interface_levels_available_res.dart';
import '../../classes/Responses/reader_read_res.dart';
import '../../classes/classes.dart';
import '../../classes/notify.dart';
import '../../constants.dart';
import '../../utility/socket_util.dart';
import 'package:get/get.dart';
import 'package:uuid/uuid.dart';

class CUPPS {
  CUPPS._();

  static final CUPPS _instance = CUPPS._();

  factory CUPPS() {
    return _instance;
  }

  late SocketService socketService;

  String airline = "44";
  String? evenToken;
  String? applicationToken;
  String? deviceToken;
  String applicationVersion = "1.0.1.0";
  PlatformDevices platformDevices = PlatformDevices.empty();

  void Function()? notifier;
  void Function(bool)? onPlatformConnectionChange;
  void Function()? onPlatformDisconnect;

  void Function(List<BcData> readData)? bcListener;
  Future<BgResponse> Function(BgData readData)? bgListener;
  void Function(List<MsData> readData)? msListener;
  void Function(List<OcData> readData)? ocListener;

  static setBcListener(void Function(List<BcData> result) listener) {
    _instance.bcListener = listener;
  }

  static setBgListener(Future<BgResponse> Function(BgData result) listener) {
    _instance.bgListener = listener;
  }

  static setMsListener(void Function(List<MsData> result) listener) {
    _instance.msListener = listener;
  }

  static setOcListener(void Function(List<OcData> result) listener) {
    _instance.ocListener = listener;
  }

  static onPlatformClose() {
    // disconnectFromPlatform();
    for (var element in _instance.platformDevices.bpDevices) {
      element.disableRetry();
    }
    for (var element in _instance.platformDevices.btDevices) {
      element.disableRetry();
    }
    for (var element in _instance.platformDevices.bcDevices) {
      element.disableRetry();
    }
    for (var element in _instance.platformDevices.ocDevices) {
      element.disableRetry();
    }
    for (var element in _instance.platformDevices.msDevices) {
      element.disableRetry();
    }
    for (var element in _instance.platformDevices.bgDevices) {
      element.disableRetry();
    }
    for (var element in _instance.platformDevices.ddDevices) {
      element.disableRetry();
    }
    for (var element in _instance.platformDevices.prDevices) {
      element.disableRetry();
    }
    for (var element in _instance.platformDevices.ziDevices) {
      element.disableRetry();
    }
    for (var element in _instance.platformDevices.zlDevices) {
      element.disableRetry();
    }


    _instance.onPlatformConnectionChange!(false);

    _instance.platformDevices = PlatformDevices.empty();
    _instance.notifier!();
    _instance.deviceToken = null;
    _instance.evenToken = null;
    _instance.applicationToken = null;
    log("Platform Disconnected");
  }

  static generateToken() {
    var uuid = const Uuid();
    String appToken = uuid.v1();
    String eToken = appToken.replaceAll("-", "").substring(0, 16);
    _instance.evenToken = eToken;
    _instance.applicationToken = appToken;
    log('Token is SET');
  }

  static int get currentMessageID => _instance.socketService.messageID;

  List<DeviceType> requiredDevices = [
    DeviceType.bp,
    DeviceType.bt,
    DeviceType.bc,
    DeviceType.ms,
    DeviceType.oc,
    DeviceType.pr,
    DeviceType.bg,
  ];

  static void setDeviceTypes(List<DeviceType> deviceTypes) {
    _instance.requiredDevices = deviceTypes;
    log("Device Type : ${deviceTypes.length}");
  }

  static Future<bool> connectToPlatform({
    required String ip,
    required int port,
    required Function() notifier,
    required void Function(bool) onPlatformConnectionChange,
    required void Function() onPlatformDisconnect,
    required String airlineCode,
    required String applicationVersion,
    bool showErrorToast = false
  }) async {
    _instance.onPlatformConnectionChange = onPlatformConnectionChange;
    _instance.onPlatformDisconnect = onPlatformDisconnect;
    _instance.notifier = notifier;
    _instance.airline = airlineCode;
    _instance.applicationVersion = applicationVersion;
    try {
      _instance.socketService = SocketService();
      bool? connection = await _instance.socketService.initialize(
          onLost: onPlatformDisconnect,
          ip: ip,
          port: port,
          rh: cuppsResponseHandler,
          onClose: onPlatformClose,
          logID: 'Platform',
          logSuff: 'Platform');
      log("Platform Socket Connection $connection");
      _instance.onPlatformConnectionChange!(connection??false);
      generateToken();
      return connection??false;
    } catch (e) {
      if(showErrorToast) {
        Get.snackbar("Platform Connection Error", e.toString());
      }else{
        log("Platform Connection Error:\n $e");
      }
      _instance.onPlatformConnectionChange!(false);
      return false;
    }
  }

  static Future<bool> authenticate() async {
    // try {
      var req = InterfaceLevelsAvailableRequest(hsXsdVersion: '01.01.0128', messageID: currentMessageID);
      String interfaceAvailableLevelsRes = await send(xml: req.toXML());
      InterfaceLevelsAvailableResponse res = InterfaceLevelsAvailableResponse.fromXML(interfaceAvailableLevelsRes);
      if (res.result.toLowerCase() == "ok") {
        if (!res.interfaceLevels.any((element) => element.level == "01.03")) {
          Get.snackbar("Authenticate Failed", "Interface Level 01.03 is Not Available!");
          return false;
        }
        var req = InterfaceLevelRequest(
            level: res.interfaceLevels.firstWhere((element) => element.level == "01.03").level, messageID: currentMessageID);
        String interfaceLevelRes = await send(xml: req.toXML());
        InterfaceLevelResponse ilRes = InterfaceLevelResponse.fromXML(interfaceLevelRes);

        if (ilRes.result.toLowerCase() == "ok") {
          AuthenticateRequest authReq = AuthenticateRequest(
            airline: _instance.airline,
            eventToken: _instance.evenToken!,
            platformDefinedParameter: _instance.applicationToken!,
            applications: [Application(applicationName: "Mobile", applicationVersion: "${_instance.applicationVersion}")],
            messageID: currentMessageID,
          );

          String authResXML = await send(xml: authReq.toXML());
          AuthenticateResponse authRes = AuthenticateResponse.fromXML(authResXML);
          log("Authenticated Token: ${authRes.deviceToken}");
          log("Total Devices Found: ${authRes.deviceList.length}");
          log(authRes.deviceList.map((e) => "${e.deviceIndex} -> ${e.deviceName} ${e.deviceParameterType} \n").join());
          _instance.deviceToken = authRes.deviceToken;
          _instance.platformDevices = authRes.platformDevices;
          _instance.notifier!();
          // connectToAvailableNeededDevices();
          return true;
        } else {
          return false;
        }
      } else {
        return false;
      }
    // } catch (e) {
    //   log("Auth Exception"+e.toString());
    //   return false;
    // }
  }

  static Future<String> send({required String xml}) async {
    return _instance.socketService.sendSocketMessage(xml);
  }

  static Future<void> respond({required String xml}) async {
    return _instance.socketService.respond(xml);
  }

  static cuppsResponseHandler(CuppsBase response, String rawXML) async {
    switch (response.messageName) {
      case CuppsMessageID.applicationStopCommandResponse:
        {
          ApplicationStopCommandResponse res = ApplicationStopCommandResponse.fromXML(rawXML);
          log(res.toXML());
          break;
        }
      case CuppsMessageID.applicationStopCommandRequest:
        {
          ApplicationStopCommandRequest req = ApplicationStopCommandRequest.fromXML(rawXML);
          // if (req.canDefer == "true") {
          ApplicationStopCommandResponse stopRes = ApplicationStopCommandResponse(
              messageID: req.messageID, messageName: "applicationStopCommandResponse", result: "OK");
          await respond(xml: stopRes.toXML());
          // }
          disconnectFromPlatform();
          break;
        }
      case CuppsMessageID.notify:
        {
          Notify res = Notify.fromXML(rawXML);
          switch (res.notificationType) {
            case NotificationType.deviceStatus:
              {
                // DeviceLockRequest req = DeviceLockRequest(messageID: currentMessageID, lockMethod: "byConnection",renew: "false");
                // send(xml: req.toXML());
                log("RECEIVED Device Status Notification");
                break;
              }
            default:
              {
                log("${res.notificationType.name} Notification Received on Platform");
              }
          }

          break;
        }
      case CuppsMessageID.bye:
        {
          ByeResponse res = ByeResponse.fromXML(rawXML);
          log("Bye Response Received ${res.result}");
          // closeSocket();
          break;
        }
      default:
        {
          log(rawXML);
        }
    }
    _instance.notifier!();
  }

  static Future<bool> connectToAvailableNeededDevices() async {
    PlatformDevices platformDevices = _instance.platformDevices;
    log("Initializing devices");
    try {
      if (_instance.requiredDevices.contains(DeviceType.bp)) {
        for (var dev in platformDevices.bpDevices) {
          dev.connect().then((connectionRes) {
            if (connectionRes.status) {
              dev.initialize().then((initRes) {
                log("$initRes");
              });
            }
          });
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.bt)) {
        for (var dev in platformDevices.btDevices) {
          dev.connect().then((connectionRes) {
            if (connectionRes.status) {
              dev.initialize().then((initRes) {
                log("$initRes");
              });
            }
          });
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.bc)) {
        for (var dev in platformDevices.bcDevices) {
          dev.connect().then((connectionRes) {
            if (connectionRes.status) {
              dev.initialize().then((initRes) {
                log("$initRes");
                if (initRes.status) {
                  dev.lock();
                }
              });
            }
          });
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.bg)) {
        for (var dev in platformDevices.bgDevices) {
          dev.connect().then((connectionRes) {
            if (connectionRes.status) {
              dev.initialize().then((initRes) {
                log("$initRes");
              });
            }
          });
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.oc)) {
        for (var dev in platformDevices.ocDevices) {
          dev.connect().then((connectionRes) {
            if (connectionRes.status) {
              dev.initialize().then((initRes) {
                log("$initRes");
                if (initRes.status) {
                  dev.lock();
                }
              });
            }
          });
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.pr)) {
        for (var dev in platformDevices.prDevices) {
          dev.connect().then((connectionRes) {
            if (connectionRes.status) {
              dev.initialize().then((initRes) {
                log("$initRes");
                if (initRes.status) {
                  dev.lock();
                }
              });
            }
          });
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.ms)) {
        for (var dev in platformDevices.msDevices) {
          dev.connect().then((connectionRes) {
            if (connectionRes.status) {
              dev.initialize().then((initRes) {
                log("$initRes");
                if (initRes.status) {
                  dev.lock();
                }
              });
            }
          });
        }
      }

      // if(platformDevices.bpDevices.isNotEmpty){
      //   CuppsCommandResponse connectRes = await platformDevices.bpDevices.first.connect();
      //   if(connectRes.status){
      //     CuppsCommandResponse initRes = await platformDevices.bpDevices.first.initialize();
      //     log("$initRes");
      //   }
      // }
      // if(platformDevices.btDevices.isNotEmpty){
      //   CuppsCommandResponse connectRes = await platformDevices.btDevices.first.connect();
      //   if(connectRes.status){
      //     CuppsCommandResponse initRes = await platformDevices.btDevices.first.initialize();
      //     log("$initRes");
      //   }
      // }
      // if(platformDevices.bcDevices.isNotEmpty){
      //   CuppsCommandResponse connectRes = await platformDevices.bcDevices.first.connect();
      //   if(connectRes.status){
      //     CuppsCommandResponse initRes = await platformDevices.bcDevices.first.initialize();
      //     log("$initRes");
      //   }
      // }
      // if(platformDevices.bgDevices.isNotEmpty){
      //   CuppsCommandResponse connectRes = await platformDevices.bgDevices.first.connect();
      //   if(connectRes.status){
      //     CuppsCommandResponse initRes = await platformDevices.bgDevices.first.initialize();
      //     log("$initRes");
      //   }
      // }
      // if(platformDevices.msDevices.isNotEmpty){
      //   CuppsCommandResponse connectRes = await platformDevices.msDevices.first.connect();
      //   if(connectRes.status){
      //     CuppsCommandResponse initRes = await platformDevices.msDevices.first.initialize();
      //     log("$initRes");
      //   }
      // }
      // if(platformDevices.prDevices.isNotEmpty){
      //   CuppsCommandResponse connectRes = await platformDevices.prDevices.first.connect();
      //   if(connectRes.status){
      //     CuppsCommandResponse initRes = await platformDevices.prDevices.first.initialize();
      //     log("$initRes");
      //   }
      // }
      // if(platformDevices.ocDevices.isNotEmpty){
      //   CuppsCommandResponse connectRes = await platformDevices.ocDevices.first.connect();
      //   if(connectRes.status){
      //     CuppsCommandResponse initRes = await platformDevices.ocDevices.first.initialize();
      //     log("$initRes");
      //   }
      // }

      // for (var dev in platformDevices.bpDevices) {
      //   CuppsCommandResponse connectRes = await dev.connect();
      //   log("$connectRes");
      //   if (connectRes.status) {
      //     CuppsCommandResponse initRes = await dev.initialize();
      //     log("$initRes");
      //   }
      // }
      // for (var dev in platformDevices.btDevices) {
      //   CuppsCommandResponse connectRes = await dev.connect();
      //   log("$connectRes");
      //   if (connectRes.status) {
      //     CuppsCommandResponse initRes = await dev.initialize();
      //     log("$initRes");
      //   }
      // }
      // for (var dev in platformDevices.bcDevices) {
      //   CuppsCommandResponse connectRes = await dev.connect();
      //   log("$connectRes");
      //   if (connectRes.status) {
      //     CuppsCommandResponse initRes = await dev.initialize();
      //     log("$initRes");
      //   }
      // }
      // for (var dev in platformDevices.beDevices) {
      //   CuppsCommandResponse connectRes = await dev.connect();
      //   log("$connectRes");
      //   if (connectRes.status) {
      //     CuppsCommandResponse initRes = await dev.initialize();
      //     log("$initRes");
      //   }
      // }
      // for (var dev in platformDevices.bgDevices) {
      //   CuppsCommandResponse connectRes = await dev.connect();
      //   log("$connectRes");
      //   if (connectRes.status) {
      //     CuppsCommandResponse initRes = await dev.initialize();
      //     log("$initRes");
      //   }
      // }
      // for (var dev in platformDevices.ddDevices) {
      //   CuppsCommandResponse connectRes = await dev.connect();
      //   log("$connectRes");
      //   if (connectRes.status) {
      //     CuppsCommandResponse initRes = await dev.initialize();
      //     log("$initRes");
      //   }
      // }
      // for (var dev in platformDevices.msDevices) {
      //   CuppsCommandResponse connectRes = await dev.connect();
      //   log("$connectRes");
      //   if (connectRes.status) {
      //     CuppsCommandResponse initRes = await dev.initialize();
      //     log("$initRes");
      //   }
      // }
      // for (var dev in platformDevices.ocDevices) {
      //   CuppsCommandResponse connectRes = await dev.connect();
      //   log("$connectRes");
      //   if (connectRes.status) {
      //     CuppsCommandResponse initRes = await dev.initialize();
      //     log("$initRes");
      //   }
      // }
      // for (var dev in platformDevices.prDevices) {
      //   CuppsCommandResponse connectRes = await dev.connect();
      //   log("$connectRes");
      //   if (connectRes.status) {
      //     CuppsCommandResponse initRes = await dev.initialize();
      //     log("$initRes");
      //   }
      // }
      // for (var dev in platformDevices.ziDevices) {
      //   CuppsCommandResponse connectRes = await dev.connect();
      //   log("$connectRes");
      //   if (connectRes.status) {
      //     CuppsCommandResponse initRes = await dev.initialize();
      //     log("$initRes");
      //   }
      // }
      // for (var dev in platformDevices.zlDevices) {
      //   CuppsCommandResponse connectRes = await dev.connect();
      //   log("$connectRes");
      //   if (connectRes.status) {
      //     CuppsCommandResponse initRes = await dev.initialize();
      //     log("$initRes");
      //   }
      // }
      log("Initializing devices done");
      return true;
    } catch (e) {
      log("$e");
      return false;
    }
  }

  static Future<void> disconnectFromPlatform() async {
    _instance.socketService.disableRetry();
    const int timeoutInSec = 10;
    try {
      if (_instance.requiredDevices.contains(DeviceType.bp)) {
        for (var dev in _instance.platformDevices.bpDevices) {
          final releaseRes = await dev.release().timeout(const Duration(seconds: timeoutInSec));
          log("$releaseRes");
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.bt)) {
        for (var dev in _instance.platformDevices.btDevices) {
          final releaseRes = await dev.release().timeout(const Duration(seconds: timeoutInSec));
          log("$releaseRes");
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.bc)) {
        for (var dev in _instance.platformDevices.bcDevices) {
          final releaseRes = await dev.release().timeout(const Duration(seconds: timeoutInSec));
          log("$releaseRes");
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.be)) {
        for (var dev in _instance.platformDevices.beDevices) {
          final releaseRes = await dev.release().timeout(const Duration(seconds: timeoutInSec));
          log("$releaseRes");
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.bg)) {
        for (var dev in _instance.platformDevices.bgDevices) {
          final releaseRes = await dev.release().timeout(const Duration(seconds: timeoutInSec));
          log("$releaseRes");
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.ms)) {
        for (var dev in _instance.platformDevices.msDevices) {
          final releaseRes = await dev.release().timeout(const Duration(seconds: timeoutInSec));
          log("$releaseRes");
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.oc)) {
        for (var dev in _instance.platformDevices.ocDevices) {
          final releaseRes = await dev.release().timeout(const Duration(seconds: timeoutInSec));
          log("$releaseRes");
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.pr)) {
        for (var dev in _instance.platformDevices.prDevices) {
          final releaseRes = await dev.release().timeout(const Duration(seconds: timeoutInSec));
          log("$releaseRes");
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.zi)) {
        for (var dev in _instance.platformDevices.ziDevices) {
          final releaseRes = await dev.release().timeout(const Duration(seconds: timeoutInSec));
          log("$releaseRes");
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.zl)) {
        for (var dev in _instance.platformDevices.zlDevices) {
          final releaseRes = await dev.release().timeout(const Duration(seconds: timeoutInSec));
          log("$releaseRes");
        }
      }
      if (_instance.requiredDevices.contains(DeviceType.dd)) {
        for (var dev in _instance.platformDevices.ddDevices) {
          final releaseRes = await dev.release().timeout(const Duration(seconds: timeoutInSec));
          log("$releaseRes");
        }
      }

      // for (var dev in _instance.platformDevices.ziDevices) {
      //   final releaseRes = await dev.release().timeout(const Duration(seconds: timeoutInSec));
      //   log("$releaseRes");
      // }
      // for (var dev in _instance.platformDevices.zlDevices) {
      //   final releaseRes = await dev.release().timeout(const Duration(seconds: timeoutInSec));
      //   log("$releaseRes");
      // }
    } catch (e) {
      log("$e");
    }
    log("Sending Bye Request");

    ByeRequest req = ByeRequest(messageID: currentMessageID);
    try {
      String? byeResponseXML = await send(xml: req.toXML()).timeout(const Duration(seconds: 5), onTimeout: () {
        _instance.socketService.disconnect();
        throw CuppsCommandException("No Response to Bye from Paltform");
      });
      ByeResponse byeRes = ByeResponse.fromXML(byeResponseXML);
      if (byeRes.result.toLowerCase() == "ok") {
        _instance.socketService.disconnect();
      } else {
        log("No Response to Bye from platform");
        _instance.socketService.disconnect();
      }
    } catch (e) {
      _instance.socketService.disconnect();
    }
  }

  static void removeDevice(String deviceIndex) async {
    await Future.delayed(const Duration(milliseconds: 100));
    _instance.platformDevices.removeDevice(deviceIndex);
    _instance.notifier;
  }

  static void setConnectionChecker(
      {Duration timeout = const Duration(seconds: 12),
      Duration interval = const Duration(seconds: 45),
      required Function retryFunction}) {
    _instance.socketService.setConnectionChecker(timeout: timeout, interval: interval, retryFunction: retryFunction);
  }
}
